import argparse
import time
from pathlib import Path

import yaml

from .obj import Obj


class ConfigLoader:
    def __init__(self):
        self._cfg = dict()
        self._args = None

    def run(self):
        args = self._parse_args()
        self._cfg.update(self._load_yaml(args.config))
        for key, value in vars(args).items():
            if key in self._cfg:
                try:
                    value = type(self._cfg[key])(value)
                except (ValueError, TypeError):
                    pass
            self._cfg[key] = value
        return Obj(**self._cfg)

    def _parse_args(self):
        parser = argparse.ArgumentParser(allow_abbrev=False)
        parser.add_argument("--config", type=str, default="config/default.yaml")
        parser.add_argument("--backup", action="store_true", default=False)
        parser.add_argument("--name", default=time.strftime("%y%m%d%H%M%S", time.localtime()), type=str)

        args, unknown_args = parser.parse_known_args()
        unknown_dict = {}
        for arg in unknown_args:
            if arg.startswith("--"):
                if "=" in arg:
                    key, value = arg.split("=", 1)
                    unknown_dict[key.lstrip("--")] = value
                else:
                    unknown_dict[arg.lstrip("--")] = True

        for key, value in unknown_dict.items():
            setattr(args, key, value)

        return args

    def _load_yaml(self, path):
        if not Path(path).exists():
            return dict()
            raise FileNotFoundError(f"Config file {path} not found")
        with open(path, "r") as f:
            return yaml.load(f, Loader=yaml.FullLoader)



import copy
from collections.abc import MutableMapping
from types import SimpleNamespace
from typing import Any, Iterator


class Obj(SimpleNamespace, MutableMapping):

    def __init__(self, **kwargs: Any):
        super().__init__()
        self._convert_and_set(kwargs)

    def _convert_value(self, value: Any) -> Any:
        if isinstance(value, dict):
            return Obj(**value)
        if isinstance(value, list):
            return [self._convert_value(v) for v in value]
        return value

    def _convert_and_set(self, data: dict) -> None:
        for k, v in data.items():
            self.__dict__[k] = self._convert_value(v)

    def __getitem__(self, key: str) -> Any:
        return self.__dict__[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.__dict__[key] = self._convert_value(value)

    def __delitem__(self, key: str) -> None:
        del self.__dict__[key]

    def __iter__(self) -> Iterator:
        return iter(self.__dict__)

    def __len__(self) -> int:
        return len(self.__dict__)

    def __getattr__(self, name: str) -> Any:
        if name.startswith("__") and name.endswith("__"):
            raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{name}'")
        return None

    def to_dict(self) -> dict:
        def _convert(obj: Any) -> Any:
            if isinstance(obj, Obj):
                return {k: _convert(v) for k, v in obj.items()}
            if isinstance(obj, list):
                return [_convert(v) for v in obj]
            return obj

        return _convert(self)

    @classmethod
    def from_dict(cls, data: dict) -> "Obj":
        return cls(**data)

    def __str__(self) -> str:
        return self.__repr__()

    def keys(self):
        return self.__dict__.keys()

    def __contains__(self, key: object) -> bool:
        return key in self.__dict__

    def to(self, device: str) -> "Obj":
        import torch

        def _move(obj: Any) -> Any:
            if isinstance(obj, torch.Tensor):
                return obj.to(device)
            if isinstance(obj, Obj):
                return obj.to(device)
            if isinstance(obj, list):
                return [_move(v) for v in obj]
            return obj

        for k, v in self.items():
            self[k] = _move(v)
        return self

    def __deepcopy__(self, memo: dict) -> "Obj":
        new_obj = self.__class__()
        memo[id(self)] = new_obj
        for k, v in self.__dict__.items():
            new_obj.__dict__[k] = copy.deepcopy(v, memo)
        return new_obj



import os
import shutil
import stat
import subprocess
from pathlib import Path

import yaml
from pathspec import GitIgnoreSpec


class BackupSystem:

    def __init__(self, cfg):
        self.cfg = cfg
        log_path = cfg.path.log if cfg.path and cfg.path.log else f"log"
        self.backup_dir = Path(log_path) / cfg.name / "backup"

    def run(self):
        if self.backup_dir.exists():
            shutil.rmtree(str(self.backup_dir))
        self.backup_dir.mkdir(parents=True, exist_ok=True)
        self._copy_project_files()
        self._save_current_config()
        self._make_backup_readonly()

    def _load_gitignore(self) -> GitIgnoreSpec:
        gitignore_path = Path(".gitignore")
        if not gitignore_path.exists():
            return GitIgnoreSpec([])
        with gitignore_path.open("r") as f:
            return GitIgnoreSpec.from_lines(f)

    def _should_ignore(self, path: Path, spec: GitIgnoreSpec) -> bool:
        git_style_path = path.relative_to(Path.cwd()).as_posix()
        if path.is_dir():
            git_style_path += "/"
        return spec.match_file(git_style_path)

    def _copy_project_files(self):
        ignore_spec = self._load_gitignore()
        backup_abs = self.backup_dir.resolve()

        for item in Path.cwd().rglob("*"):
            item_abs = item.resolve()
            if item_abs == backup_abs or backup_abs in item_abs.parents:
                continue
            if self._should_ignore(item, ignore_spec):
                continue
            if item.is_dir():
                continue
            dest = self.backup_dir / item.relative_to(Path.cwd())
            dest.parent.mkdir(parents=True, exist_ok=True)
            shutil.copy2(item, dest)

    def _save_current_config(self):
        config_path = self.backup_dir / self.cfg.config
        config_path.parent.mkdir(exist_ok=True)
        with config_path.open("w") as f:
            yaml.dump(self.cfg.to_dict(), f)

    def _make_backup_readonly(self):
        for root, dirs, files in os.walk(self.backup_dir):
            for name in dirs + files:
                path = Path(root) / name
                if not path.is_dir():
                    path.chmod(0o444)
